@@ -1,7 +1,4 @@
 use crate::types::{Asn1Readable, SimpleAsn1Readable, Tlv};
-use alloc::borrow::Cow;
-use alloc::string::ToString;
-use alloc::vec::Vec;
 use core::fmt;
 
 /// ParseError are returned when there is an error parsing the ASN.1 data.
@@ -21,6 +18,9 @@ pub enum ParseErrorKind {
     InvalidSetOrdering,
     /// An OPTIONAL DEFAULT was written with a default value.
     EncodedDefault,
+    /// OID value is longer than the maximum size rust-asn1 can store. This is
+    /// a limitation of rust-asn1.
+    OidTooLong,
 }
 
 #[derive(Debug, PartialEq)]
@@ -47,6 +47,7 @@ impl ParseError {
     }
 
     #[doc(hidden)]
+    #[must_use]
     pub fn add_location(mut self, loc: ParseLocation) -> Self {
         if (self.parse_depth as usize) < self.parse_locations.len() {
             self.parse_locations[self.parse_depth as usize] = Some(loc);
@@ -56,27 +57,72 @@ impl ParseError {
     }
 }
 
+#[cfg(feature = "std")]
+impl std::error::Error for ParseError {}
+
+// Wraps an `Option<T>`, but `fmt::Debug` will only render `Some` values and
+// panics on others.
+struct SomeFmtOption<T>(Option<T>);
+
+impl<T: fmt::Debug> fmt::Debug for SomeFmtOption<T> {
+    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
+        self.0.as_ref().unwrap().fmt(f)
+    }
+}
+
 impl fmt::Debug for ParseError {
     fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
         let mut f = f.debug_struct("ParseError");
         f.field("kind", &self.kind);
         if self.parse_depth > 0 {
-            f.field(
-                "location",
-                &self.parse_locations[..self.parse_depth as usize]
-                    .iter()
-                    .rev()
-                    .map(|l| match l.as_ref().unwrap() {
-                        ParseLocation::Field(f) => Cow::<'_, str>::Borrowed(f),
-                        ParseLocation::Index(i) => Cow::<'_, str>::Owned(i.to_string()),
-                    })
-                    .collect::<Vec<_>>(),
-            );
+            let mut locations = [
+                SomeFmtOption(None),
+                SomeFmtOption(None),
+                SomeFmtOption(None),
+                SomeFmtOption(None),
+                SomeFmtOption(None),
+                SomeFmtOption(None),
+                SomeFmtOption(None),
+                SomeFmtOption(None),
+            ];
+            for (i, location) in self.parse_locations[..self.parse_depth as usize]
+                .iter()
+                .rev()
+                .enumerate()
+            {
+                locations[i] = match location.as_ref().unwrap() {
+                    ParseLocation::Field(ref f) => SomeFmtOption(Some(f as &dyn fmt::Debug)),
+                    ParseLocation::Index(ref i) => SomeFmtOption(Some(i as &dyn fmt::Debug)),
+                }
+            }
+
+            f.field("location", &&locations[..self.parse_depth as usize]);
         }
         f.finish()
     }
 }
 
+impl fmt::Display for ParseError {
+    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
+        write!(f, "ASN.1 parsing error: ")?;
+        match self.kind {
+            ParseErrorKind::InvalidValue => write!(f, "invalid value"),
+            ParseErrorKind::UnexpectedTag { actual } => {
+                write!(f, "unexpected tag (got {})", actual)
+            }
+            ParseErrorKind::ShortData => write!(f, "short data"),
+            ParseErrorKind::IntegerOverflow => write!(f, "integer overflow"),
+            ParseErrorKind::ExtraData => write!(f, "extra data"),
+            ParseErrorKind::InvalidSetOrdering => write!(f, "SET value was ordered incorrectly"),
+            ParseErrorKind::EncodedDefault => write!(f, "DEFAULT value was explicitly encoded"),
+            ParseErrorKind::OidTooLong => write!(
+                f,
+                "OBJECT IDENTIFIER was too large to be stored in rust-asn1's buffer"
+            ),
+        }
+    }
+}
+
 /// The result of a `parse`. Either a successful value or a `ParseError`.
 pub type ParseResult<T> = Result<T, ParseError>;
 
@@ -203,6 +249,7 @@ impl<'a> Parser<'a> {
 
     /// Reads a single ASN.1 element from the parser. Which type you are reading is determined by
     /// the type parameter `T`.
+    #[inline]
     pub fn read_element<T: Asn1Readable<'a>>(&mut self) -> ParseResult<T> {
         T::parse(self)
     }
@@ -267,10 +314,10 @@ mod tests {
     use super::Parser;
     use crate::types::Asn1Readable;
     use crate::{
-        BMPString, BigUint, BitString, Choice1, Choice2, Choice3, Enumerated, GeneralizedTime,
-        IA5String, ObjectIdentifier, ParseError, ParseErrorKind, ParseLocation, ParseResult,
-        PrintableString, Sequence, SequenceOf, SetOf, Tlv, UniversalString, UtcTime, Utf8String,
-        VisibleString,
+        BMPString, BigInt, BigUint, BitString, Choice1, Choice2, Choice3, Enumerated,
+        GeneralizedTime, IA5String, ObjectIdentifier, ParseError, ParseErrorKind, ParseLocation,
+        ParseResult, PrintableString, Sequence, SequenceOf, SetOf, Tlv, UniversalString, UtcTime,
+        Utf8String, VisibleString,
     };
     #[cfg(feature = "const-generics")]
     use crate::{Explicit, Implicit};
@@ -325,6 +372,12 @@ mod tests {
                     .add_location(ParseLocation::Field("Abc::123")),
                 "ParseError { kind: InvalidValue, location: [\"Abc::123\"] }",
             ),
+            (
+                ParseError::new(ParseErrorKind::InvalidValue)
+                    .add_location(ParseLocation::Index(12))
+                    .add_location(ParseLocation::Field("Abc::123")),
+                "ParseError { kind: InvalidValue, location: [\"Abc::123\", 12] }",
+            ),
         ]
         .iter()
         {
@@ -332,6 +385,31 @@ mod tests {
         }
     }
 
+    #[test]
+    fn test_parse_error_display() {
+        for (e, expected) in [
+            (
+                ParseError::new(ParseErrorKind::InvalidValue),
+                "ASN.1 parsing error: invalid value",
+            ),
+            (
+                ParseError::new(ParseErrorKind::ShortData)
+                    .add_location(ParseLocation::Field("Abc::123")),
+                "ASN.1 parsing error: short data",
+            ),
+            (
+                ParseError::new(ParseErrorKind::UnexpectedTag { actual: 12 })
+                    .add_location(ParseLocation::Index(12))
+                    .add_location(ParseLocation::Field("Abc::123")),
+                "ASN.1 parsing error: unexpected tag (got 12)",
+            ),
+        ]
+        .iter()
+        {
+            assert_eq!(&format!("{}", e), expected)
+        }
+    }
+
     fn assert_parses_cb<
         'a,
         T: fmt::Debug + PartialEq,
@@ -533,6 +611,48 @@ mod tests {
         ]);
     }
 
+    #[test]
+    fn test_parse_int_i32() {
+        assert_parses::<i32>(&[
+            (Ok(0), b"\x02\x01\x00"),
+            (Ok(127), b"\x02\x01\x7f"),
+            (Ok(128), b"\x02\x02\x00\x80"),
+            (Ok(256), b"\x02\x02\x01\x00"),
+            (Ok(-128), b"\x02\x01\x80"),
+            (Ok(-129), b"\x02\x02\xff\x7f"),
+            (Ok(-256), b"\x02\x02\xff\x00"),
+            (Ok(core::i32::MAX), b"\x02\x04\x7f\xff\xff\xff"),
+            (
+                Err(ParseError::new(ParseErrorKind::UnexpectedTag {
+                    actual: 0x3,
+                })),
+                b"\x03\x00",
+            ),
+            (
+                Err(ParseError::new(ParseErrorKind::ShortData)),
+                b"\x02\x02\x00",
+            ),
+            (Err(ParseError::new(ParseErrorKind::ShortData)), b""),
+            (Err(ParseError::new(ParseErrorKind::ShortData)), b"\x02"),
+            (
+                Err(ParseError::new(ParseErrorKind::IntegerOverflow)),
+                b"\x02\x09\x02\x00\x00\x00\x00\x00\x00\x00\x00",
+            ),
+            (
+                Err(ParseError::new(ParseErrorKind::InvalidValue)),
+                b"\x02\x05\x00\x00\x00\x00\x01",
+            ),
+            (
+                Err(ParseError::new(ParseErrorKind::InvalidValue)),
+                b"\x02\x02\xff\x80",
+            ),
+            (
+                Err(ParseError::new(ParseErrorKind::InvalidValue)),
+                b"\x02\x00",
+            ),
+        ])
+    }
+
     #[test]
     fn test_parse_int_i8() {
         assert_parses::<i8>(&[
@@ -591,9 +711,33 @@ mod tests {
         ]);
     }
 
+    #[test]
+    fn test_parse_bigint() {
+        assert_parses::<BigInt>(&[
+            (Ok(BigInt::new(b"\x80").unwrap()), b"\x02\x01\x80"),
+            (Ok(BigInt::new(b"\xff").unwrap()), b"\x02\x01\xff"),
+            (
+                Ok(BigInt::new(b"\x00\xff\xff").unwrap()),
+                b"\x02\x03\x00\xff\xff",
+            ),
+            (
+                Err(ParseError::new(ParseErrorKind::InvalidValue)),
+                b"\x02\x02\xff\xff",
+            ),
+            (
+                Ok(BigInt::new(b"\xff\x7f\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff").unwrap()),
+                b"\x02\x0c\xff\x7f\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff",
+            ),
+            (
+                Err(ParseError::new(ParseErrorKind::InvalidValue)),
+                b"\x02\x00",
+            ),
+        ]);
+    }
+
     #[test]
     fn test_parse_object_identifier() {
-        assert_parses::<ObjectIdentifier<'_>>(&[
+        assert_parses::<ObjectIdentifier>(&[
             (
                 Ok(ObjectIdentifier::from_string("2.5").unwrap()),
                 b"\x06\x01\x55",
