@@ -61,6 +61,7 @@ pub trait SimpleAsn1Writable<'a>: Sized {
 }
 
 impl<'a, T: SimpleAsn1Writable<'a>> Asn1Writable<'a> for T {
+    #[inline]
     fn write(&self, w: &mut Writer) {
         w.write_tlv(Self::TAG, move |dest| self.write_data(dest));
     }
@@ -105,15 +106,18 @@ impl<'a> Tlv<'a> {
 }
 
 impl<'a> Asn1Readable<'a> for Tlv<'a> {
+    #[inline]
     fn parse(parser: &mut Parser<'a>) -> ParseResult<Self> {
         parser.read_tlv()
     }
 
+    #[inline]
     fn can_parse(_tag: u8) -> bool {
         true
     }
 }
 impl<'a> Asn1Writable<'a> for Tlv<'a> {
+    #[inline]
     fn write(&self, w: &mut Writer) {
         w.write_tlv(self.tag, move |dest| dest.extend_from_slice(self.data))
     }
@@ -542,7 +546,8 @@ macro_rules! impl_asn1_element_for_int {
                 }
 
                 for i in (1..num_bytes + 1).rev() {
-                    dest.push((self >> ((i - 1) * 8)) as u8);
+                    let digit = self.checked_shr((i - 1) * 8).unwrap_or(0);
+                    dest.push(digit as u8);
                 }
             }
         }
@@ -551,7 +556,7 @@ macro_rules! impl_asn1_element_for_int {
 
 impl_asn1_element_for_int!(i8; true);
 impl_asn1_element_for_int!(u8; false);
-impl_asn1_element_for_int!(i32; false);
+impl_asn1_element_for_int!(i32; true);
 impl_asn1_element_for_int!(u32; false);
 impl_asn1_element_for_int!(i64; true);
 impl_asn1_element_for_int!(u64; false);
@@ -593,17 +598,53 @@ impl<'a> SimpleAsn1Writable<'a> for BigUint<'a> {
     }
 }
 
-impl<'a> SimpleAsn1Readable<'a> for ObjectIdentifier<'a> {
-    const TAG: u8 = 0x06;
-    fn parse_data(data: &'a [u8]) -> ParseResult<ObjectIdentifier<'a>> {
-        ObjectIdentifier::from_der(data)
-            .ok_or_else(|| ParseError::new(ParseErrorKind::InvalidValue))
+/// Arbitrary sized signed integer. Contents may be accessed as `&[u8]` of
+/// big-endian data. Its contents always match the DER encoding of a value
+/// (i.e. they are minimal)
+#[derive(PartialEq, Clone, Copy, Debug, Hash)]
+pub struct BigInt<'a> {
+    data: &'a [u8],
+}
+
+impl<'a> BigInt<'a> {
+    /// Create a new BigInt from already encoded data. `data` must be encoded
+    /// as required by DER: minimally and if the high bit would be set in the
+    /// first octet, a leading \x00 should be prepended (to disambiguate from
+    /// negative values).
+    pub fn new(data: &'a [u8]) -> Option<Self> {
+        validate_integer(data, true).ok()?;
+        Some(BigInt { data })
+    }
+
+    /// Returns the contents of the integer as big-endian bytes.
+    pub fn as_bytes(&self) -> &'a [u8] {
+        self.data
     }
 }
-impl<'a> SimpleAsn1Writable<'a> for ObjectIdentifier<'a> {
+
+impl<'a> SimpleAsn1Readable<'a> for BigInt<'a> {
+    const TAG: u8 = 0x02;
+    fn parse_data(data: &'a [u8]) -> ParseResult<Self> {
+        BigInt::new(data).ok_or_else(|| ParseError::new(ParseErrorKind::InvalidValue))
+    }
+}
+impl<'a> SimpleAsn1Writable<'a> for BigInt<'a> {
+    const TAG: u8 = 0x02;
+    fn write_data(&self, dest: &mut Vec<u8>) {
+        dest.extend_from_slice(self.data);
+    }
+}
+
+impl<'a> SimpleAsn1Readable<'a> for ObjectIdentifier {
+    const TAG: u8 = 0x06;
+    fn parse_data(data: &'a [u8]) -> ParseResult<ObjectIdentifier> {
+        ObjectIdentifier::from_der(data)
+    }
+}
+impl<'a> SimpleAsn1Writable<'a> for ObjectIdentifier {
     const TAG: u8 = 0x06;
     fn write_data(&self, dest: &mut Vec<u8>) {
-        dest.extend_from_slice(&self.der_encoded);
+        dest.extend_from_slice(self.as_der());
     }
 }
 
@@ -801,12 +842,14 @@ impl<'a, T: Asn1Readable<'a>> Asn1Readable<'a> for Option<T> {
         }
     }
 
+    #[inline]
     fn can_parse(tag: u8) -> bool {
         T::can_parse(tag)
     }
 }
 
 impl<'a, T: Asn1Writable<'a>> Asn1Writable<'a> for Option<T> {
+    #[inline]
     fn write(&self, w: &mut Writer) {
         if let Some(v) = self {
             w.write_element(v);
