@@ -34,6 +34,7 @@ impl Writer<'_> {
     }
 
     /// Writes a single element to the output.
+    #[inline]
     pub fn write_element<'a, T: Asn1Writable<'a>>(&mut self, val: &T) {
         val.write(self);
     }
@@ -132,7 +133,7 @@ mod tests {
     use super::{_insert_at_position, write, write_single, Writer};
     use crate::types::Asn1Writable;
     use crate::{
-        parse_single, BMPString, BigUint, BitString, Choice1, Choice2, Choice3, Enumerated,
+        parse_single, BMPString, BigInt, BigUint, BitString, Choice1, Choice2, Choice3, Enumerated,
         GeneralizedTime, IA5String, ObjectIdentifier, OwnedBitString, PrintableString, Sequence,
         SequenceOf, SequenceOfWriter, SequenceWriter, SetOf, SetOfWriter, Tlv, UniversalString,
         UtcTime, Utf8String, VisibleString,
@@ -267,6 +268,29 @@ mod tests {
         ]);
     }
 
+    #[test]
+    fn test_write_i32() {
+        assert_writes::<i32>(&[
+            (0, b"\x02\x01\x00"),
+            (127, b"\x02\x01\x7f"),
+            (128, b"\x02\x02\x00\x80"),
+            (255, b"\x02\x02\x00\xff"),
+            (256, b"\x02\x02\x01\x00"),
+            (-1, b"\x02\x01\xff"),
+            (-128, b"\x02\x01\x80"),
+            (-129, b"\x02\x02\xff\x7f"),
+        ]);
+    }
+
+    #[test]
+    fn test_write_u8() {
+        assert_writes::<u8>(&[
+            (0, b"\x02\x01\x00"),
+            (127, b"\x02\x01\x7f"),
+            (128, b"\x02\x02\x00\x80"),
+        ]);
+    }
+
     #[test]
     fn test_write_i8() {
         assert_writes::<i8>(&[
@@ -288,6 +312,17 @@ mod tests {
         ]);
     }
 
+    #[test]
+    fn test_write_bigint() {
+        assert_writes::<BigInt>(&[
+            (BigInt::new(b"\xff").unwrap(), b"\x02\x01\xff"),
+            (
+                BigInt::new(b"\xff\x7f\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff").unwrap(),
+                b"\x02\x0c\xff\x7f\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff",
+            ),
+        ]);
+    }
+
     #[test]
     fn test_write_object_identifier() {
         assert_writes::<ObjectIdentifier>(&[
